"""
Models for notifications app.
"""

import json
import jsonpickle
import six

from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string


@six.python_2_unicode_compatible
class SentNotification(models.Model):
    """
    Stores info on the notification that was sent.
    """

    STATUS_PENDING = 0
    STATUS_SUCCESS = 1
    STATUS_FAILED = 2
    STATUS_USER_DISABLED = 3

    STATUSES = (
        (0, 'Pending'),
        (1, 'Success'),
        (2, 'Failed'),
        (3, 'User Disabled')
    )

    text_content = models.TextField(null=True, blank=True)
    html_content = models.TextField(null=True, blank=True)
    sent_from = models.CharField(max_length=100, null=True, blank=True)
    recipients = models.CharField(max_length=2000)  # Comma separated list of emails or numbers
    subject = models.CharField(max_length=255, null=True, blank=True)
    extra_data = models.TextField(null=True, blank=True)  # json dictionary
    date_sent = models.DateTimeField()
    status = models.PositiveSmallIntegerField(choices=STATUSES, default=STATUS_PENDING)
    notification_class = models.CharField(max_length=255)
    error_message = models.TextField(null=True, blank=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, default=None, null=True, on_delete=models.SET_NULL)
    attachments = models.TextField(null=True, blank=True)

    def __str__(self):
        return self.notification_class

    def get_recipients(self):
        """
        Return the list of recipients for the notification. Recipient is defined by the notification class.
        """

        return self.recipients.split(',')

    def resend(self):
        """
        Re-sends the notification by calling the notification class' resend method
        """

        notification_class = import_string(self.notification_class)
        notification_class.resend(self)

    def get_extra_data(self):
        """
        Return extra data that was saved
        """

        if not self.extra_data:
            return {}
        else:
            return json.loads(self.extra_data)

    def get_attachments(self):
        if self.attachments:
            return jsonpickle.loads(self.attachments)
        else:
            return None


@six.python_2_unicode_compatible
class Notification(models.Model):
    """
    NotificationClasses are created on app init.
    """
    notification_class = models.CharField(max_length=255, unique=True)
    verbose_name = models.CharField(max_length=255, blank=True, null=True)
    can_disable = models.BooleanField(default=True)

    def __str__(self):
        return self.verbose_name if self.verbose_name else self.notification_class


class UserNotification(models.Model):
    """
    Add a User Notification record, then add disabled notifications to disable records.
    On your user Admin, add the field user_notification
    """
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        primary_key=True
    )
    disabled_notifications = models.ManyToManyField(Notification)



from django.test import TestCase, Client


class ViewsTests(TestCase):
    def test_index(self):
        client = Client()
        response = client.get('/herald/')
        self.assertContains(response, 'MyNotification')

    def test_preview_text(self):
        client = Client()
        response = client.get('/herald/0/text/')
        self.assertContains(response, 'Hello World')

    def test_preview_html(self):
        client = Client()
        response = client.get('/herald/0/html/')
        self.assertContains(response, '<html><body>Hello World</body></html>')



"""
Testing custom commands
"""
from datetime import timedelta, datetime

from django.core.management import call_command
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone
from django.utils.six import StringIO

from herald.models import SentNotification
from herald.management.commands.delnotifs import valid_date

MSG = 'Successfully deleted {num} notification(s)'
NOTIFICATION_CLASS = 'tests.notifications.MyNotification'




class DeleteNotificationNoArgs(TestCase):

    def test_date_validator(self):
        self.assertEqual(
            valid_date('2017-01-01'),
            datetime(2017,1,1)
        )

    def test_delete_today(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()
        ).save()
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()+timedelta(days=1)
        ).save()
        out = StringIO()
        call_command('delnotifs', stdout=out)
        self.assertIn(MSG.format(num=1), out.getvalue())

    def test_do_not_delete_tomorrow(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()+timedelta(days=1)
        ).save()
        out = StringIO()
        call_command('delnotifs', stdout=out)
        self.assertIn(MSG.format(num=0), out.getvalue())

    def test_do_not_delete_yesterday(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now() - timedelta(days=1)
        ).save()
        out = StringIO()
        call_command('delnotifs', stdout=out)
        self.assertIn(MSG.format(num=0), out.getvalue())


class DeleteNotificationArgs(TestCase):

    def test_delete_today_start_arg(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()
        ).save()
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()+timedelta(days=1)
        ).save()
        out = StringIO()
        today = timezone.now().replace(hour=0, minute=0, second=0, microsecond=0)
        call_command('delnotifs', stdout=out, start=str(today))
        self.assertIn(MSG.format(num=1), out.getvalue())

    def test_do_not_delete_tomorrow_end_arg(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now()+timedelta(days=1)
        ).save()
        out = StringIO()
        two_days_later = timezone.now().replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(days=2)
        call_command('delnotifs', stdout=out, end=str(two_days_later))
        self.assertIn(MSG.format(num=1), out.getvalue())

    def test_do_accept_bad_args(self):
        SentNotification(
            notification_class=NOTIFICATION_CLASS,
            date_sent=timezone.now() - timedelta(days=1)
        ).save()
        out = StringIO()

        with self.assertRaises(ValidationError):
            call_command('delnotifs', stdout=out, start='blargh')

        with self.assertRaises(ValidationError):
            call_command('delnotifs', stdout=out, start='01-01-2016')



"""
Herald notifications for working with django.contrib.auth
"""

from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.models import Site
from django.template import loader
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.urls import reverse

from ... import registry
from ...base import EmailNotification


class PasswordResetEmail(EmailNotification):
    """
    Email sent when requesting password reset using forgot password feature.
    This replaces django's default email
    """
    template_name = 'password_reset'

    def __init__(self, user, site_name=None, domain=None, extra_email_context=None, use_https=False,
                 token_generator=default_token_generator,
                 subject_template_name='registration/password_reset_subject.txt',
                 email_template_name='registration/password_reset_email.html', html_email_template_name=None):
        self.to_emails = [user.email]
        self.site_name = site_name
        self.domain = domain
        self.user = user
        self.token_generator = token_generator
        self.use_https = use_https
        self.extra_email_context = extra_email_context
        self.subject_template_name = subject_template_name
        self.email_template_name = email_template_name
        self.html_email_template_name = html_email_template_name

    def get_context_data(self):
        context = super(PasswordResetEmail, self).get_context_data()

        if not self.site_name or self.domain:
            current_site = Site.objects.get_current()
            self.site_name = current_site.name
            self.domain = current_site.domain

        protocol = 'https' if self.use_https else 'http'
        uid = urlsafe_base64_encode(force_bytes(self.user.pk))
        token = self.token_generator.make_token(self.user)

        context.update({
            'full_reset_url': '{}://{}{}'.format(
                protocol,
                self.domain,
                reverse('password_reset_confirm', kwargs={'uidb64': uid, 'token': token})
            ),
            'email': self.user.email,
            'domain': self.domain,
            'site_name': self.site_name,
            'uid': uid,
            'user': self.user,
            'token': token,
            'protocol': protocol,
            'template_name': self.email_template_name,
            'html_template_name': self.html_email_template_name,
        })

        if self.extra_email_context is not None:
            context.update(self.extra_email_context)

        return context

    def get_subject(self):
        subject = super(PasswordResetEmail, self).get_subject()

        if not subject:
            # subject was not defined on the class. Use the default subject template to get the subject.
            subject = loader.render_to_string(self.subject_template_name, self.get_context_data())
            # can't have newlines
            return ''.join(subject.splitlines())

        return subject

    @staticmethod
    def get_demo_args():
        User = get_user_model()
        return [User(**{User.USERNAME_FIELD: 'username@example.com'})]


registry.register(PasswordResetEmail)



from django.test import TestCase

from herald import registry
from herald.base import EmailNotification


class InitTests(TestCase):
    def test_register(self):
        class TestNotification(EmailNotification):
            pass

        registry.register(TestNotification)

        self.assertEqual(len(registry._registry), 5)

        registry.unregister(TestNotification)

        self.assertEqual(len(registry._registry), 4)

    def test_register_decorator(self):
        @registry.register_decorator()
        class TestNotification(EmailNotification):
            pass

        self.assertEqual(len(registry._registry), 5)

        registry.unregister(TestNotification)

        self.assertEqual(len(registry._registry), 4)

    def test_register_invalid(self):
        class TestNotification(object):
            pass

        with self.assertRaises(ValueError):
            registry.register(TestNotification)

        self.assertEqual(len(registry._registry), 4)



"""
Notification classes. Used for sending texts and emails
"""
__version__ = '0.2'

default_app_config = 'herald.apps.HeraldConfig'


class NotificationRegistry(object):
    """
    Stores the notification classes that get registered.
    """

    def __init__(self):
        self._registry = []

    def register(self, kls):
        """
        Register a notification class
        """

        from .base import NotificationBase

        if not issubclass(kls, NotificationBase):
                raise ValueError('Notification must subclass NotificationBase.')

        self._registry.append(kls)

    def unregister(self, kls):
        """
        Unregister a notification class
        """

        self._registry.remove(kls)

    def register_decorator(self):
        """
        Registers the given notification with Django Herold

        """

        def _notification_wrapper(kls):

            self.register(kls)

            return kls
        return _notification_wrapper


registry = NotificationRegistry()  # pylint: disable=C0103


def autodiscover():
    """
    Auto discover notification registrations in any file called "notifications" in any app.
    """
    from django.utils.module_loading import autodiscover_modules

    autodiscover_modules('notifications', register_to=registry)



"""
Admin for notifications
"""

from functools import update_wrapper

from django.conf.urls import url
from django.contrib import admin, messages
from django.contrib.admin.options import csrf_protect_m
from django.contrib.admin.utils import unquote
from django.utils.safestring import mark_safe
from django.urls import reverse

from .models import SentNotification, Notification


@admin.register(SentNotification)
class SentNotificationAdmin(admin.ModelAdmin):
    """
    Admin for viewing historical notifications sent
    """

    list_display = ('notification_class', 'recipients', 'subject', 'date_sent', 'status')
    list_filter = ('status', 'notification_class')
    date_hierarchy = 'date_sent'
    readonly_fields = ('resend', )
    search_fields = ('recipients', 'subject', 'text_content', 'html_content', 'sent_from')

    def resend(self, obj):
        """
        Creates a link field that takes user to re-send view to resend the notification
        """

        opts = self.model._meta  # pylint: disable=W0212
        resend_url = reverse(
            'admin:%s_%s_resend' % (opts.app_label, opts.model_name),
            current_app=self.admin_site.name,
            args=(obj.pk, )
        )

        return mark_safe('<a href="{}">Resend</a>'.format(resend_url))

    def get_urls(self):
        urls = super(SentNotificationAdmin, self).get_urls()
        opts = self.model._meta  # pylint: disable=W0212

        def wrap(view):
            """
            Copied from super class
            """
            def wrapper(*args, **kwargs):
                """
                Copied from super class
                """
                return self.admin_site.admin_view(view)(*args, **kwargs)
            return update_wrapper(wrapper, view)

        info = opts.app_label, opts.model_name

        return [
            url(r'^(.+)/resend/$', wrap(self.resend_view), name='%s_%s_resend' % info),
        ] + urls

    @csrf_protect_m
    def resend_view(self, request, object_id, extra_context=None):  # pylint: disable=W0613
        """
        View that re-sends the notification
        """

        obj = self.get_object(request, unquote(object_id))

        success = obj.resend()

        if success:
            self.message_user(request, 'The notification was resent successfully.', messages.SUCCESS)
        else:
            self.message_user(request, 'The notification failed to resend.', messages.ERROR)

        return self.response_post_save_change(request, obj)


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    """
    Admin for viewing/managing notifications in the system
    """
    list_display = ('notification_class', 'verbose_name', 'can_disable')
    search_fields = ('notification_class', 'verbose_name')




# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-09 14:06
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('herald', '0003_auto_20170830_1617'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sentnotification',
            name='error_message',
            field=models.TextField(blank=True, null=True),
        ),
    ]



import datetime

from django.utils import timezone
from django.core.management.base import BaseCommand

from ...models import SentNotification


def valid_date(s):
    return datetime.datetime.strptime(s, "%Y-%m-%d")


class Command(BaseCommand):
    help = 'Deletes notifications between the date ranges specified.'

    def add_arguments(self, parser):
        parser.add_argument('--start', help="includes this date, format YYYY-MM-DD", type=valid_date)
        parser.add_argument('--end', help="up to this date, format YYYY-MM-DD", type=valid_date)

    def handle(self, *args, **options):
        today = timezone.now().replace(hour=0, minute=0, second=0, microsecond=0)
        start_date = options['start'] if options['start'] else today
        end_date = options['end'] if options['end'] else today + datetime.timedelta(days=1)
        qs = SentNotification.objects.filter(date_sent__range=[start_date, end_date])
        present_notifications = qs.count()
        deleted_notifications = qs.delete()
        deleted_num = deleted_notifications[0] if deleted_notifications is not None else present_notifications
        self.stdout.write('Successfully deleted {num} notification(s)'.format(num=deleted_num))



# Generated by Django 2.0.5 on 2018-05-16 22:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('herald', '0004_auto_20171009_0906'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='notification_class',
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='verbose_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SentNotification',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('text_content', models.TextField(null=True, blank=True)),
                ('html_content', models.TextField(null=True, blank=True)),
                ('sent_from', models.CharField(max_length=100, null=True, blank=True)),
                ('recipients', models.CharField(max_length=2000)),
                ('subject', models.CharField(max_length=255, null=True, blank=True)),
                ('extra_data', models.TextField(null=True, blank=True)),
                ('date_sent', models.DateTimeField()),
                ('status', models.PositiveSmallIntegerField(default=0, choices=[(0, 'Pending'), (1, 'Success'), (2, 'Failed')])),
                ('notification_class', models.CharField(max_length=255)),
                ('error_message', models.CharField(max_length=255, null=True, blank=True)),
            ],
        ),
    ]



from email.mime.image import MIMEImage

from herald import registry
from herald.base import EmailNotification
from herald.base import TwilioTextNotification


class MyNotification(EmailNotification):
    context = {'hello': 'world'}
    template_name = 'hello_world'
    to_emails = ['test@test.com']

    def get_attachments(self):
        # this returns two attachments, one a text file, the other an inline attachment that can be referred to in a
        # template using cid: notation
        fp = open('tests/python.jpeg', 'rb')
        img = MIMEImage(fp.read())
        img.add_header('Content-ID', '<{}>'.format('python.jpg'))

        raw_data = 'Some Report Data'

        return [
            ('Report.txt', raw_data, 'text/plain'),
            img,
        ]

registry.register(MyNotification)


class MyOtherNotification(EmailNotification):
    context = {'hello': 'world'}
    template_name = 'hello_world'
    to_emails = ['test@test.com']

registry.register(MyOtherNotification)


class MyTwilioNotification(TwilioTextNotification):
    context = {'hello': 'world'}
    template_name = 'hello_world'
    to_emails = ['test@test.com']


registry.register(MyTwilioNotification)



from django.contrib.auth import get_user_model
from django.core import mail
from django.test import TestCase

from herald.contrib.auth.forms import HeraldPasswordResetForm


class ContribAuthTests(TestCase):
    def test_save_form(self):
        User = get_user_model()
        User.objects.create_user(username='test@example.com', email='test@example.com', password='password')
        form = HeraldPasswordResetForm({'email': 'test@example.com'})
        form.is_valid()
        form.save()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEquals(mail.outbox[0].to, ['test@example.com'])
        
    def test_save_form_domain_override(self):
        User = get_user_model()
        User.objects.create_user(username='test@example.com', email='test@example.com', password='password')
        form = HeraldPasswordResetForm({'email': 'test@example.com'})
        form.is_valid()
        form.save(domain_override='foo')
        self.assertEqual(len(mail.outbox), 1)
        self.assertEquals(mail.outbox[0].to, ['test@example.com'])



"""
Django app config for herald. Using this to call autodiscover
"""

from django.apps import AppConfig
from django.db.utils import OperationalError, ProgrammingError


class HeraldConfig(AppConfig):
    """
    Django app config for herald. Using this to call autodiscover
    """

    name = 'herald'

    def ready(self):
        from herald import registry

        self.module.autodiscover()

        Notification = self.get_model('Notification')

        try:
            # add any new notifications to database.
            for index, klass in enumerate(registry._registry):
                notification, created = Notification.objects.get_or_create(
                    notification_class=klass.get_class_path(),
                    defaults={
                        'verbose_name': klass.get_verbose_name(),
                        'can_disable': klass.can_disable,
                    }
                )

                if not created:
                    notification.verbose_name = klass.get_verbose_name()
                    notification.can_disable = klass.can_disable
                    notification.save()

        except OperationalError:
            # if the table is not created yet, just keep going.
            pass
        except ProgrammingError:
            # if the database is not created yet, keep going (ie: during testing)
            pass



"""
Base notification classes
"""

import json
from email.mime.base import MIMEBase
from mimetypes import guess_type

import jsonpickle
import re
import six

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.mail import EmailMultiAlternatives
from django.template import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.utils import timezone
from django.core.files import File

from .models import SentNotification


class NotificationBase(object):
    """
    base class for sending notifications
    """

    render_types = []
    template_name = None
    context = None
    user = None
    can_disable = True
    verbose_name = None

    def get_context_data(self):
        """
        :return: the context data for rendering the email or text template
        """

        context = self.context or {}

        if settings.DEBUG:
            context['base_url'] = ''
        else:
            site = Site.objects.get_current()
            context['base_url'] = 'http://' + site.domain

        return context

    @classmethod
    def get_verbose_name(cls):
        if cls.verbose_name:
            return cls.verbose_name
        else:
            return re.sub(
                r'((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))',
                r' \1',
                cls.__name__
            )

    @classmethod
    def get_class_path(cls):
        return '{}.{}'.format(cls.__module__, cls.__name__)

    def send(self, raise_exception=False, user=None):
        """
        Handles the preparing the notification for sending. Called to trigger the send from code.
        If raise_exception is True, it will raise any exceptions rather than simply logging them.
        returns boolean whether or not the notification was sent successfully
        """
        context = self.get_context_data()

        recipients = self.get_recipients()

        if 'text' in self.render_types:
            text_content = self.render('text', context)
        else:
            text_content = None

        if 'html' in self.render_types:
            html_content = self.render('html', context)
        else:
            html_content = None

        sent_from = self.get_sent_from()
        subject = self.get_subject()
        extra_data = self.get_extra_data()

        sent_notification = SentNotification(
            recipients=','.join(recipients),
            text_content=text_content,
            html_content=html_content,
            sent_from=sent_from,
            subject=subject,
            extra_data=json.dumps(extra_data) if extra_data else None,
            notification_class=self.get_class_path(),
            attachments=self._get_encoded_attachments(),
            user=user,
        )

        return self.resend(sent_notification, raise_exception=raise_exception)

    def _get_encoded_attachments(self):
        attachments = self.get_attachments()

        new_attachments = []

        for attachment in attachments or []:
            if isinstance(attachment, File):
                attachment.seek(0)
                new_attachments.append((attachment.name, attachment.read(), guess_type(attachment.name)[0]))
            else:
                new_attachments.append(attachment)

        return jsonpickle.dumps(new_attachments)

    def get_recipients(self):
        """
        Returns a list of recipients. However the subclass defines that. (emails, phone numbers, etc)
        """

        raise NotImplementedError('Must implement get_recipients.')

    def get_extra_data(self):
        """
        Returns a dictionary of extra data to be stored, and used for sending.
        MUST BE JSON SERIALIZABLE
        """

        return {}

    def get_sent_from(self):
        """
        Returns a "sent from" string. However the subclass defines that. (email, phone number, etc)
        """

        raise NotImplementedError('Must implement get_recipients.')

    def get_subject(self):
        """
        Returns a subject string. Optional.
        """

        return None

    def get_attachments(self):
        """
        Return a list of attachments or None.

        This only works with email.
        """
        return None

    def render(self, render_type, context):
        """
        Renders the template

        :param render_type: the content type to render
        :param context: context data dictionary
        :return: the rendered content
        """

        assert render_type in self.render_types, 'Invalid Render Type'

        try:
            content = render_to_string('herald/{}/{}.{}'.format(
                render_type,
                self.template_name,
                'txt' if render_type == 'text' else render_type
            ), context)
        except TemplateDoesNotExist:
            content = None

            if settings.DEBUG:
                raise

        return content

    @staticmethod
    def get_demo_args():
        """
        Returns iterable of arguments needed to initialize notification for demo purposes
        Usually you want to generate dummy data here for testing
        """
        return []

    @classmethod
    def resend(cls, sent_notification, raise_exception=False):
        """
        Takes a saved sent_notification and sends it again.
        returns boolean whether or not the notification was sent successfully
        """

        # handle skipping a notification based on user preference
        if hasattr(sent_notification.user, 'usernotification'):
            notifications = sent_notification.user.usernotification
            if notifications.disabled_notifications.filter(notification_class=cls.get_class_path()).exists():
                sent_notification.date_sent = timezone.now()
                sent_notification.status = sent_notification.STATUS_USER_DISABLED
                sent_notification.save()
                return True

        try:
            cls._send(
                sent_notification.get_recipients(),
                sent_notification.text_content,
                sent_notification.html_content,
                sent_notification.sent_from,
                sent_notification.subject,
                sent_notification.get_extra_data(),
                sent_notification.get_attachments(),
            )
            sent_notification.status = sent_notification.STATUS_SUCCESS
        except Exception as exc:  # pylint: disable=W0703
            # we want to handle any exception whatsoever
            sent_notification.status = sent_notification.STATUS_FAILED
            sent_notification.error_message = six.text_type(exc)

            if raise_exception:
                raise exc

        sent_notification.date_sent = timezone.now()
        sent_notification.save()

        return sent_notification.status == sent_notification.STATUS_SUCCESS

    @staticmethod
    def _send(recipients, text_content=None, html_content=None, sent_from=None, subject=None, extra_data=None,
              attachments=None):
        """
        Handles the actual sending of the notification. Sub classes must override this
        """

        raise NotImplementedError('Must implement send.')


class EmailNotification(NotificationBase):
    """
    Base class for email notifications
    """

    render_types = ['text', 'html']
    from_email = None
    subject = None
    to_emails = None
    bcc = None
    cc = None  # pylint: disable=C0103
    headers = None
    reply_to = None
    attachments = None

    def get_context_data(self):
        context = super(EmailNotification, self).get_context_data()
        context['subject'] = self.subject
        return context

    def get_recipients(self):
        return self.to_emails

    def get_sent_from(self):
        from_email = self.from_email
        if not from_email:
            from_email = settings.DEFAULT_FROM_EMAIL

        return from_email

    def get_subject(self):
        return self.subject

    def get_extra_data(self):
        extra_data = {}

        if self.bcc:
            extra_data['bcc'] = self.bcc

        if self.cc:
            extra_data['cc'] = self.cc

        if self.headers:
            extra_data['headers'] = self.headers

        if self.reply_to:
            extra_data['reply_to'] = self.reply_to

        return extra_data

    def get_attachments(self):
        """
        Return a list of attachments or None.

        This only works with email.
        """
        return self.attachments

    @staticmethod
    def _send(recipients, text_content=None, html_content=None, sent_from=None, subject=None, extra_data=None,
              attachments=None):

        extra_data = extra_data or {}

        mail = EmailMultiAlternatives(
            subject=subject,
            body=text_content,
            from_email=sent_from,
            to=recipients,
            bcc=extra_data.get('bcc', None),
            headers=extra_data.get('headers', None),
            cc=extra_data.get('cc', None),
            reply_to=extra_data.get('reply_to', None),
        )

        if html_content:
            mail.attach_alternative(html_content, 'text/html')

        for attachment in (attachments or []):
            # All mimebase attachments must have a Content-ID or Content-Disposition header
            # or they will show up as unnamed attachments"
            if isinstance(attachment, MIMEBase):
                if attachment.get('Content-ID', False):
                    # if you are sending attachment with content id,
                    # subtype must be 'related'.
                    mail.mixed_subtype = 'related'

                mail.attach(attachment)
            else:
                mail.attach(*attachment)

        mail.send()


class TwilioTextNotification(NotificationBase):
    """
    Base class for text notifications.
    Uses twilio
    """

    render_types = ['text']
    from_number = None
    to_number = None

    def get_recipients(self):
        return [self.to_number]

    def get_sent_from(self):
        from_number = self.from_number
        if not from_number:
            try:
                from_number = settings.TWILIO_DEFAULT_FROM_NUMBER
            except AttributeError:
                raise Exception(
                    'TWILIO_DEFAULT_FROM_NUMBER setting is required for sending a TwilioTextNotification'
                )

        return from_number

    @staticmethod
    def _send(recipients, text_content=None, html_content=None, sent_from=None, subject=None, extra_data=None,
              attachments=None):
        try:
            # twilio version 6
            from twilio.rest import Client
        except ImportError:
            try:
                # twillio version < 6
                from twilio.rest import TwilioRestClient as Client
            except ImportError:
                raise Exception('Twilio is required for sending a TwilioTextNotification.')

        try:
            account_sid = settings.TWILIO_ACCOUNT_SID
            auth_token = settings.TWILIO_AUTH_TOKEN
        except AttributeError:
            raise Exception(
                'TWILIO_ACCOUNT_SID and TWILIO_AUTH_TOKEN settings are required for sending a TwilioTextNotification'
            )

        client = Client(account_sid, auth_token)

        client.messages.create(
            body=text_content,
            to=recipients[0],
            from_=sent_from
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-08-30 21:17
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('herald', '0002_sentnotification_attachments'),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('notification_class', models.CharField(max_length=80, unique=True)),
                ('verbose_name', models.CharField(blank=True, max_length=100, null=True)),
                ('can_disable', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='UserNotification',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to=settings.AUTH_USER_MODEL)),
                ('disabled_notifications', models.ManyToManyField(to='herald.Notification')),
            ],
        ),
        migrations.AddField(
            model_name='sentnotification',
            name='user',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='sentnotification',
            name='status',
            field=models.PositiveSmallIntegerField(choices=[(0, 'Pending'), (1, 'Success'), (2, 'Failed'), (3, 'User Disabled')], default=0),
        ),
    ]



from django.contrib.auth import get_user_model
from django.test import TestCase, Client
from django.utils import timezone
from django.urls import reverse

from mock import patch

from herald.models import SentNotification


class ViewsTests(TestCase):
    def setUp(self):
        get_user_model().objects.create_superuser('admin', 'admin@example.com', 'password')
        self.client = Client()
        self.client.login(username='admin', password='password')
        self.notification = SentNotification.objects.create(
            recipients='test@example.com',
            date_sent=timezone.now(),
            status=SentNotification.STATUS_SUCCESS,
            notification_class='tests.notifications.MyNotification'
        )

    def test_index(self):
        response = self.client.get(reverse('admin:herald_sentnotification_changelist'))
        self.assertEqual(response.status_code, 200)

    def test_detail(self):
        response = self.client.get(reverse('admin:herald_sentnotification_change', args=(self.notification.pk, )))
        self.assertEqual(response.status_code, 200)

    def test_resend(self):
        with patch.object(SentNotification, 'resend') as mocked_resend:
            response = self.client.get(
                reverse('admin:herald_sentnotification_resend', args=(self.notification.pk,)),
                follow=True
            )
            mocked_resend.assert_called_once()

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            'The notification was resent successfully.',
            [m.message for m in list(response.context['messages'])]
        )

    def test_resend_fail(self):
        with patch.object(SentNotification, 'resend') as mocked_resend:
            mocked_resend.return_value = False
            response = self.client.get(
                reverse('admin:herald_sentnotification_resend', args=(self.notification.pk,)),
                follow=True
            )
            mocked_resend.assert_called_once()

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            'The notification failed to resend.',
            [m.message for m in list(response.context['messages'])]
        )



import django

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.sites',
    'django.contrib.admin',
    'django.contrib.sessions',
    'django.contrib.messages',
    'herald',
    'tests',
    'herald.contrib.auth',
)

MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

SITE_ID = 1

ROOT_URLCONF = 'tests.urls'

DEBUG = True

USE_TZ = True

SECRET_KEY = 'foobar'

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'APP_DIRS': True,
    'OPTIONS': {
        'context_processors': [
            "django.contrib.auth.context_processors.auth",
            "django.template.context_processors.debug",
            "django.template.context_processors.i18n",
            "django.template.context_processors.media",
            "django.template.context_processors.static",
            "django.template.context_processors.tz",
            "django.contrib.messages.context_processors.messages"
        ]
    }
}]



"""
Form classes to be used with django.contrib.auth
"""

from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site

from .notifications import PasswordResetEmail


class HeraldPasswordResetForm(PasswordResetForm):
    """
    Form used when entering your email to send a password reset email
    """

    def save(self, domain_override=None, subject_template_name='registration/password_reset_subject.txt',
             email_template_name='registration/password_reset_email.html', use_https=False,
             token_generator=default_token_generator, from_email=None, request=None, html_email_template_name=None,
             extra_email_context=None):
        """
        Generates a one-use only link for resetting password and sends to the
        user.
        """

        email = self.cleaned_data["email"]
        for user in self.get_users(email):
            if not domain_override:
                current_site = get_current_site(request)
                site_name = current_site.name
                domain = current_site.domain
            else:
                site_name = domain = domain_override

            PasswordResetEmail(
                user,
                site_name=site_name,
                domain=domain,
                extra_email_context=extra_email_context,
                use_https=use_https,
                token_generator=token_generator,
                subject_template_name=subject_template_name,
                email_template_name=email_template_name,
                html_email_template_name=html_email_template_name
            ).send()



from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib import admin
from django.contrib.auth import urls as auth_urls
from django.conf.urls import url, include


urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url('^', include(auth_urls)),
    url(r'^herald/', include('herald.urls')),
]



"""
Urls for herald app
"""

from django.conf.urls import url

from .views import TestNotificationList, TestNotification

urlpatterns = [
    url(r'^$', TestNotificationList.as_view(), name='herald_preview_list'),
    url(r'^(?P<index>\d+)/(?P<type>[\w\-]+)/$', TestNotification.as_view(), name='herald_preview'),
]


import os
from setuptools import find_packages, setup

VERSION = __import__('herald').__version__


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ''


install_requires = ['django>=1.8', 'six', 'jsonpickle']

dev_requires = ['pytz']

setup(
    name='django-herald',
    version=VERSION,
    author='Worthwhile',
    author_email='devs@worthwhile.com',
    install_requires=install_requires,
    extra_require={
        'dev': install_requires + dev_requires,
    },
    packages=find_packages(include=('herald', 'herald.*')),
    include_package_data=True,  # declarations in MANIFEST.in
    license='MIT',
    url='https://github.com/worthwhile/django-herald/',
    download_url='https://github.com/worthwhile/django-herald/tarball/'+VERSION,
    description="Django library for separating the message content from transmission method",
    long_description=read_file('README.md'),
    keywords=['django', 'notifications', 'messaging'],
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: MIT License',
    ],
)



import jsonpickle
from django.core import mail
from django.core.files import File
from django.core.mail import EmailMultiAlternatives
from django.template import TemplateDoesNotExist
from django.test import TestCase, override_settings
from mock import patch

try:
    # twilio version 6
    from twilio.rest.api.v2010.account import MessageList
except ImportError:
    # twillio version < 6
    from twilio.rest.resources import Messages as MessageList

from herald.base import NotificationBase, EmailNotification, TwilioTextNotification
from herald.models import SentNotification
from .notifications import MyNotification


class BaseNotificationTests(TestCase):
    def test_get_context_data(self):
        with override_settings(DEBUG=True):
            self.assertDictEqual(
                MyNotification().get_context_data(),
                {'hello': 'world', 'base_url': '', 'subject': None}
            )

        self.assertDictEqual(
            MyNotification().get_context_data(),
            {'hello': 'world', 'base_url': 'http://example.com', 'subject': None}
        )

    def test_get_recipients(self):
        self.assertRaises(NotImplementedError, NotificationBase().get_recipients)

    def test_get_extra_data(self):
        self.assertDictEqual(NotificationBase().get_extra_data(), {})

    def test_get_sent_from(self):
        self.assertRaises(NotImplementedError, NotificationBase().get_sent_from)

    def test_get_subject(self):
        self.assertIsNone(NotificationBase().get_subject())

    def test_get_demo_args(self):
        self.assertListEqual(NotificationBase.get_demo_args(), [])

    def test_private_send(self):
        self.assertRaises(NotImplementedError, NotificationBase()._send, [])

    def test_get_attachments(self):
        self.assertIsNone(NotificationBase().get_attachments())

    def test_send(self):
        with patch.object(MyNotification, 'resend') as mocked_resend:
            MyNotification().send()
            mocked_resend.assert_called_once()
            obj = mocked_resend.call_args[0][0]
            self.assertEqual(obj.recipients, 'test@test.com')

    def test_send_no_text(self):
        class DummyNotification(EmailNotification):
            render_types = ['html']
            to_emails = ['test@test.com']

        with patch.object(DummyNotification, 'resend') as mocked_resend:
            DummyNotification().send()
            mocked_resend.assert_called_once()
            obj = mocked_resend.call_args[0][0]
            self.assertEqual(obj.recipients, 'test@test.com')
            self.assertIsNone(obj.text_content)

    def test_real_send(self):
        MyNotification().send()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEquals(mail.outbox[0].to, ['test@test.com'])

    def test_real_send_attachments(self):
        MyNotification().send()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].attachments[0][1],'Some Report Data')

    def test_render_no_type(self):
        class DummyNotification(NotificationBase):
            pass

        with self.assertRaises(AssertionError):
            DummyNotification().render('text', {})

    def test_render_invalid_template(self):
        class DummyNotification(NotificationBase):
            render_types = ['text']
            template_name = 'does_not_exist'

        self.assertIsNone(DummyNotification().render('text', {}))

    @override_settings(DEBUG=True)
    def test_render_invalid_template_debug(self):
        class DummyNotification(NotificationBase):
            render_types = ['text']
            template_name = 'does_not_exist'

        with self.assertRaises(TemplateDoesNotExist):
            DummyNotification().render('text', {})

    def test_render_invalid(self):
        class DummyNotification(NotificationBase):
            render_types = ['text']
            template_name = 'hello_world'

        self.assertEqual(DummyNotification().render('text', {}), 'Hello World')

    def test_resend_error(self):
        notification = SentNotification()

        with patch.object(NotificationBase, '_send') as mocked__send:
            mocked__send.side_effect = Exception
            result = NotificationBase.resend(notification)
            self.assertFalse(result)

    def test_resend_error_raise(self):
        notification = SentNotification()

        with patch.object(NotificationBase, '_send') as mocked__send:
            mocked__send.side_effect = Exception
            self.assertRaises(Exception, NotificationBase.resend, notification, raise_exception=True)

    def test_resend(self):
        notification = SentNotification()

        with patch.object(NotificationBase, '_send') as mocked__send:
            result = NotificationBase.resend(notification)
            self.assertTrue(result)

    def test_get_verbose_name(self):
        class TestNotification(EmailNotification):
            pass

        self.assertEqual(TestNotification.get_verbose_name(), 'Test Notification')

        class TestNotification2(EmailNotification):
            verbose_name = 'A verbose name'

        self.assertEqual(TestNotification2.get_verbose_name(), 'A verbose name')

    def test_get_encoded_attachments_none(self):
        class TestNotification(EmailNotification):
            attachments = []

        self.assertJSONEqual(TestNotification()._get_encoded_attachments(), [])

    def test_get_encoded_attachments_basic(self):
        class TestNotification(EmailNotification):
            attachments = [('Report.txt', 'raw_data', 'text/plain')]

        self.assertJSONEqual(
            TestNotification()._get_encoded_attachments(),
            [{'py/tuple': ['Report.txt', 'raw_data', 'text/plain']}]
        )

    def test_get_encoded_attachments_file(self):
        class TestNotification(EmailNotification):
            attachments = [File(open('tests/python.jpeg', 'rb'))]

        attachments = jsonpickle.loads(TestNotification()._get_encoded_attachments())
        self.assertEqual(attachments[0][0], 'tests/python.jpeg')
        self.assertEqual(attachments[0][2], 'image/jpeg')


class EmailNotificationTests(TestCase):
    def test_get_recipients(self):
        self.assertListEqual(MyNotification().get_recipients(), ['test@test.com'])

    def test_get_sent_from(self):
        class TestNotification(EmailNotification):
            from_email = 'bob@example.com'

        self.assertEqual(TestNotification().get_sent_from(), 'bob@example.com')

    def test_get_sent_from_default(self):
        class TestNotification(EmailNotification):
            from_email = None

        with override_settings(DEFAULT_FROM_EMAIL='default@example.com'):
            self.assertEqual(TestNotification().get_sent_from(), 'default@example.com')

    def test_get_subject(self):
        class TestNotification(EmailNotification):
            subject = 'test subject'

        self.assertEqual(TestNotification().get_subject(), 'test subject')

    def test_get_extra_data_none(self):
        self.assertDictEqual(EmailNotification().get_extra_data(), {})

    def test_get_extra_data(self):
        class TestNotification(EmailNotification):
            bcc = 'bcc@test.com'
            cc = 'cc@test.com'
            headers = {'HEADER': 'test'}
            reply_to = 'reply_to@test.com'

        self.assertDictEqual(TestNotification().get_extra_data(), {
            'bcc': 'bcc@test.com',
            'cc': 'cc@test.com',
            'headers': {'HEADER': 'test'},
            'reply_to': 'reply_to@test.com',
        })

    def test_send_html_content(self):
        class TestNotification(EmailNotification):
            subject = 'test subject'

        with patch.object(EmailMultiAlternatives, 'attach_alternative') as mocked_attach_alternative:
            TestNotification._send([], text_content='Text')
            mocked_attach_alternative.assert_not_called()

        with patch.object(EmailMultiAlternatives, 'attach_alternative') as mocked_attach_alternative:
            TestNotification._send([], html_content='Text')
            mocked_attach_alternative.assert_called_once_with('Text', 'text/html')


class TwilioNotificationTests(TestCase):
    def test_get_recipients(self):
        class TestNotification(TwilioTextNotification):
            to_number = '1231231234'

        self.assertListEqual(TestNotification().get_recipients(), ['1231231234'])

    def test_get_sent_from(self):
        class TestNotification(TwilioTextNotification):
            from_number = '1231231234'

        self.assertEqual(TestNotification().get_sent_from(), '1231231234')

    def test_get_sent_from_default(self):
        class TestNotification(TwilioTextNotification):
            from_number = None

        with override_settings(TWILIO_DEFAULT_FROM_NUMBER='1231231234'):
            self.assertEqual(TestNotification().get_sent_from(), '1231231234')

    def test_get_sent_from_default_error(self):
        class TestNotification(TwilioTextNotification):
            from_number = None

        self.assertRaisesMessage(
            Exception,
            'TWILIO_DEFAULT_FROM_NUMBER setting is required for sending a TwilioTextNotification',
            TestNotification().get_sent_from
        )

    @override_settings(
        TWILIO_ACCOUNT_SID='sid',
        TWILIO_AUTH_TOKEN='token'
    )
    def test_send(self):
        class TestNotification(TwilioTextNotification):
            from_number = '1231231234'
            to_number = '1231231234'
            template_name = 'hello_world'

        with patch.object(MessageList, 'create') as mocked_create:
            TestNotification().send()
            mocked_create.assert_called_once_with(
                body='Hello World',
                to='1231231234',
                from_='1231231234'
            )

    def test_send_no_settings(self):
        class TestNotification(TwilioTextNotification):
            from_number = '1231231234'
            to_number = '1231231234'
            template_name = 'hello_world'

        with self.assertRaisesMessage(Exception, 'TWILIO_ACCOUNT_SID and TWILIO_AUTH_TOKEN settings are required for '
                                                 'sending a TwilioTextNotification'):
            TestNotification().send(raise_exception=True)



"""
Views for testing notifications. Should not be present in production
"""

from django.http import HttpResponse
from django.views.generic import TemplateView, View

from . import registry


class TestNotificationList(TemplateView):
    """
    View for listing out all notifications with links to view rendered versions of them
    """

    template_name = 'herald/test/notification_list.html'

    def get_context_data(self, **kwargs):
        context = super(TestNotificationList, self).get_context_data(**kwargs)

        context['notifications'] = [
            (index, x.__name__, x.render_types, (y.__name__ for y in x.__bases__)) for index, x in enumerate(registry._registry)  # pylint: disable=W0212
        ]

        return context


class TestNotification(View):
    """
    View for showing rendered test notification
    """

    def get(self, request, *args, **kwargs):  # pylint: disable=W0613
        """
        GET request
        """

        index = int(kwargs['index'])
        render_type = kwargs['type']

        obj = registry._registry[index](*registry._registry[index].get_demo_args())  # pylint: disable=W0212

        context = obj.get_context_data()

        content = obj.render(render_type, context)

        return HttpResponse(content, content_type='text/{}'.format(render_type))



import six
from mock import patch

from django.test import TestCase

from herald.models import SentNotification, Notification
from tests.notifications import MyNotification


class SentNotificationTests(TestCase):
    def test_str(self):
        notification = SentNotification(notification_class='tests.notifications.MyNotification')
        self.assertEqual(six.text_type(notification), 'tests.notifications.MyNotification')

    def test_get_recipients(self):
        notification = SentNotification(recipients='test@test.com,example@example.com')
        self.assertListEqual(notification.get_recipients(), ['test@test.com','example@example.com'])

    def test_get_extra_data_none(self):
        notification = SentNotification()
        self.assertDictEqual(notification.get_extra_data(), {})

    def test_get_extra_data(self):
        notification = SentNotification(extra_data='{"something":["one","two"]}')
        self.assertDictEqual(notification.get_extra_data(), {'something': ['one', 'two']})

    def test_resend(self):
        notification = SentNotification(notification_class='tests.notifications.MyNotification')
        with patch.object(MyNotification, 'resend') as mocked_resend:
            notification.resend()
            mocked_resend.assert_called_once_with(notification)


class NotificationTests(TestCase):
    def test_str(self):
        notification = Notification(notification_class='tests.notifications.MyNotification')
        self.assertEqual(six.text_type(notification), 'tests.notifications.MyNotification')
